import re
from collections.abc import Callable
from pathlib import Path

import boto3
from boto3 import Session

from dbt_platform_helper.constants import PLATFORM_CONFIG_FILE
from dbt_platform_helper.domain.maintenance_page import MaintenancePage
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.config_validator import ConfigValidator
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.io import ClickIOProviderException
from dbt_platform_helper.providers.vpc import Vpc
from dbt_platform_helper.providers.vpc import VpcProvider
from dbt_platform_helper.providers.vpc import VpcProviderException
from dbt_platform_helper.utils.application import Application
from dbt_platform_helper.utils.application import ApplicationNotFoundException
from dbt_platform_helper.utils.application import load_application
from dbt_platform_helper.utils.aws import get_connection_string
from dbt_platform_helper.utils.aws import wait_for_log_group_to_exist


class DatabaseCopy:
    def __init__(
        self,
        app: str,
        database: str,
        auto_approve: bool = False,
        load_application: Callable[[str], Application] = load_application,
        # TODO: DBTP-1960: We inject VpcProvider as a callable here so that it can be instantiated within the method.  To be improved
        vpc_provider: Callable[[Session], VpcProvider] = VpcProvider,
        db_connection_string: Callable[
            [Session, str, str, str, Callable], str
        ] = get_connection_string,
        maintenance_page: Callable[[str, str, list[str], str, str], None] = MaintenancePage,
        io: ClickIOProvider = ClickIOProvider(),
        config_provider: ConfigProvider = ConfigProvider(ConfigValidator()),
    ):
        self.app = app
        self.database = database
        self.auto_approve = auto_approve
        self.vpc_provider = vpc_provider
        self.db_connection_string = db_connection_string
        self.io = io
        self.config_provider = config_provider

        if not self.app:
            if not Path(PLATFORM_CONFIG_FILE).exists():
                self.io.abort_with_error(
                    "You must either be in a deploy repo, or provide the --app option."
                )

            config = self.config_provider.load_and_validate_platform_config()
            self.app = config["application"]

        try:
            self.application = load_application(self.app)
        except ApplicationNotFoundException:
            self.io.abort_with_error(f"No such application '{app}'.")

        self.maintenance_page = maintenance_page(self.application)

    def _execute_operation(self, is_dump: bool, env: str, vpc_name: str, filename: str):
        vpc_name = self.enrich_vpc_name(env, vpc_name)

        environments = self.application.environments
        environment = environments.get(env)
        if not environment:
            self.io.abort_with_error(
                f"No such environment '{env}'. Available environments are: {', '.join(environments.keys())}"
            )

        env_session = environment.session

        try:
            vpc_provider = self.vpc_provider(env_session)
            vpc_config = vpc_provider.get_vpc(self.app, env, vpc_name)
        except VpcProviderException as ex:
            self.io.abort_with_error(str(ex))

        if not vpc_config.security_groups:
            self.io.abort_with_error(f"No security groups found in vpc '{vpc_name}'")

        database_identifier = f"{self.app}-{env}-{self.database}"

        try:
            db_connection_string = self.db_connection_string(
                env_session, self.app, env, database_identifier
            )
        except Exception as exc:
            self.io.abort_with_error(f"{exc} (Database: {database_identifier})")

        try:
            task_arn = self.run_database_copy_task(
                env_session, env, vpc_config, is_dump, db_connection_string, filename
            )
        except Exception as exc:
            self.io.abort_with_error(f"{exc} (Account id: {self.account_id(env)})")

        if is_dump:
            message = f"Dumping {self.database} from the {env} environment into S3"
        else:
            message = f"Loading data into {self.database} in the {env} environment from S3"

        self.io.info(message)
        self.io.info(
            f"Task {task_arn} started. Waiting for it to complete (this may take some time)..."
        )
        self.tail_logs(is_dump, env)

    def enrich_vpc_name(self, env, vpc_name):
        if not vpc_name:
            if not Path(PLATFORM_CONFIG_FILE).exists():
                self.io.abort_with_error(
                    "You must either be in a deploy repo, or provide the vpc name option."
                )
            config = self.config_provider.load_and_validate_platform_config()
            env_config = self.config_provider.apply_environment_defaults(config)["environments"]
            vpc_name = env_config.get(env, {}).get("vpc")
        return vpc_name

    def run_database_copy_task(
        self,
        session: boto3.session.Session,
        env: str,
        vpc: Vpc,
        is_dump: bool,
        db_connection_string: str,
        filename: str,
    ) -> str:
        client = session.client("ecs")
        action = "dump" if is_dump else "load"
        dump_file_name = filename if filename else "data_dump"
        env_vars = [
            {"name": "DATA_COPY_OPERATION", "value": action.upper()},
            {"name": "DB_CONNECTION_STRING", "value": db_connection_string},
            {"name": "DUMP_FILE_NAME", "value": dump_file_name},
        ]
        if not is_dump:
            env_vars.append({"name": "ECS_CLUSTER", "value": f"{self.app}-{env}"})

        response = client.run_task(
            taskDefinition=f"arn:aws:ecs:eu-west-2:{self.account_id(env)}:task-definition/{self.app}-{env}-{self.database}-{action}",
            cluster=f"{self.app}-{env}",
            capacityProviderStrategy=[
                {"capacityProvider": "FARGATE", "weight": 1, "base": 0},
            ],
            networkConfiguration={
                "awsvpcConfiguration": {
                    "subnets": vpc.private_subnets,
                    "securityGroups": vpc.security_groups,
                    "assignPublicIp": "DISABLED",
                }
            },
            overrides={
                "containerOverrides": [
                    {
                        "name": f"{self.app}-{env}-{self.database}-{action}",
                        "environment": env_vars,
                    }
                ]
            },
        )

        return response.get("tasks", [{}])[0].get("taskArn")

    def dump(self, env: str, vpc_name: str, filename: str = None):
        self._execute_operation(True, env, vpc_name, filename)

    def load(self, env: str, vpc_name: str, filename: str = None):
        if self.is_confirmed_ready_to_load(env):
            self._execute_operation(False, env, vpc_name, filename)

    def copy(
        self,
        from_env: str,
        to_env: str,
        from_vpc: str,
        to_vpc: str,
        services: tuple[str],
        template: str,
        no_maintenance_page: bool = False,
    ):
        to_vpc = self.enrich_vpc_name(to_env, to_vpc)
        if not no_maintenance_page:
            self.maintenance_page.activate(to_env, services, template, to_vpc)
        self.dump(from_env, from_vpc, f"data_dump_{to_env}")
        self.load(to_env, to_vpc, f"data_dump_{to_env}")
        if not no_maintenance_page:
            self.maintenance_page.deactivate(to_env)

    def is_confirmed_ready_to_load(self, env: str) -> bool:
        if self.auto_approve:
            return True
        try:
            return self.io.confirm(
                f"\nWARNING: the load operation is destructive and will delete the {self.database} database in the {env} environment. Continue?"
            )
        except ClickIOProviderException:
            return False

    def tail_logs(self, is_dump: bool, env: str):
        action = "dump" if is_dump else "load"
        log_group_name = f"/ecs/{self.app}-{env}-{self.database}-{action}"
        log_group_arn = f"arn:aws:logs:eu-west-2:{self.account_id(env)}:log-group:{log_group_name}"
        self.io.warn(f"Tailing {log_group_name} logs")
        session = self.application.environments[env].session
        log_client = session.client("logs")
        wait_for_log_group_to_exist(log_client, log_group_name)
        response = log_client.start_live_tail(logGroupIdentifiers=[log_group_arn])

        stopped = False
        for data in response["responseStream"]:
            if stopped:
                break
            results = data.get("sessionUpdate", {}).get("sessionResults", [])
            for result in results:
                message = result.get("message")

                if message:
                    match = re.match(r"(Stopping|Aborting) data (load|dump).*", message)
                    if match:
                        if match.group(1) == "Aborting":
                            self.io.abort_with_error(
                                "Task aborted abnormally. See logs above for details."
                            )
                        stopped = True
                    self.io.info(message)

    def account_id(self, env):
        envs = self.application.environments
        if env in envs:
            return envs.get(env).account_id



import subprocess

from dbt_platform_helper.providers.cloudformation import CloudFormation
from dbt_platform_helper.providers.copilot import connect_to_addon_client_task
from dbt_platform_helper.providers.copilot import create_addon_client_task
from dbt_platform_helper.providers.ecs import ECS
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.secrets import Secrets
from dbt_platform_helper.utils.application import Application


class Conduit:
    def __init__(
        self,
        application: Application,
        secrets_provider: Secrets,
        cloudformation_provider: CloudFormation,
        ecs_provider: ECS,
        io: ClickIOProvider = ClickIOProvider(),
        subprocess: subprocess = subprocess,
        connect_to_addon_client_task=connect_to_addon_client_task,
        create_addon_client_task=create_addon_client_task,
    ):

        self.application = application
        self.secrets_provider = secrets_provider
        self.cloudformation_provider = cloudformation_provider
        self.ecs_provider = ecs_provider
        self.subprocess = subprocess
        self.io = io
        self.connect_to_addon_client_task = connect_to_addon_client_task
        self.create_addon_client_task = create_addon_client_task

    def start(self, env: str, addon_name: str, access: str = "read"):
        clients = self._initialise_clients(env)
        addon_type, cluster_arn, parameter_name, task_name = self._get_addon_details(
            addon_name, access
        )

        self.io.info(f"Checking if a conduit task is already running for {addon_type}")
        task_arns = self.ecs_provider.get_ecs_task_arns(cluster_arn, task_name)
        if not task_arns:
            self.io.info("Creating conduit task")
            self.create_addon_client_task(
                clients["iam"],
                clients["ssm"],
                self.subprocess,
                self.application,
                env,
                addon_type,
                addon_name,
                task_name,
                access,
            )

            self.io.info("Updating conduit task")
            self._update_stack_resources(
                self.application.name,
                env,
                addon_type,
                addon_name,
                task_name,
                parameter_name,
                access,
            )

            task_arns = self.ecs_provider.get_ecs_task_arns(cluster_arn, task_name)

        else:
            self.io.info("Conduit task already running")

        self.io.info(f"Checking if exec is available for conduit task...")

        self.ecs_provider.ecs_exec_is_available(cluster_arn, task_arns)

        self.io.info("Connecting to conduit task")
        self.connect_to_addon_client_task(
            clients["ecs"], self.subprocess, self.application.name, env, cluster_arn, task_name
        )

    def _initialise_clients(self, env):
        return {
            "ecs": self.application.environments[env].session.client("ecs"),
            "iam": self.application.environments[env].session.client("iam"),
            "ssm": self.application.environments[env].session.client("ssm"),
        }

    def _get_addon_details(self, addon_name, access):
        addon_type = self.secrets_provider.get_addon_type(addon_name)
        cluster_arn = self.ecs_provider.get_cluster_arn()
        parameter_name = self.secrets_provider.get_parameter_name(addon_type, addon_name, access)
        task_name = self.ecs_provider.get_or_create_task_name(addon_name, parameter_name)

        return addon_type, cluster_arn, parameter_name, task_name

    def _update_stack_resources(
        self,
        app_name,
        env,
        addon_type,
        addon_name,
        task_name,
        parameter_name,
        access,
    ):
        self.cloudformation_provider.add_stack_delete_policy_to_task_role(task_name)
        stack_name = self.cloudformation_provider.update_conduit_stack_resources(
            app_name,
            env,
            addon_type,
            addon_name,
            task_name,
            parameter_name,
            access,
        )
        self.io.info("Waiting for conduit task update to complete...")
        self.cloudformation_provider.wait_for_cloudformation_to_reach_status(
            "stack_update_complete", stack_name
        )



import boto3


class Redis:

    def __init__(self, client: boto3.client):
        self.client = client
        self.engine = "redis"

    def get_reference(self) -> str:
        return self.engine.lower()

    def get_supported_versions(self) -> list[str]:
        supported_versions_response = self.client.describe_cache_engine_versions(Engine=self.engine)

        supported_versions = [
            version["EngineVersion"]
            for version in supported_versions_response["CacheEngineVersions"]
        ]

        return supported_versions



import re
import subprocess
from abc import ABC
from abc import abstractmethod
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version
from typing import Union

from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.semantic_version import SemanticVersion


def set_up_retry():
    session = Session()
    retries = Retry(total=5, backoff_factor=0.1, status_forcelist=[403, 500, 502, 503, 504])
    session.mount("https://", HTTPAdapter(max_retries=retries))
    return session


class InstalledVersionProviderException(PlatformException):
    pass


class InstalledToolNotFoundException(InstalledVersionProviderException):
    def __init__(
        self,
        tool_name: str,
    ):
        super().__init__(f"Package '{tool_name}' not found.")


class VersionProvider(ABC):
    @abstractmethod
    def get_semantic_version() -> SemanticVersion:
        raise NotImplementedError("Must be implemented in subclasses")


class InstalledVersionProvider:
    @staticmethod
    def get_semantic_version(tool_name: str) -> SemanticVersion:
        try:
            return SemanticVersion.from_string(version(tool_name))

        except PackageNotFoundError:
            raise InstalledToolNotFoundException(tool_name)


class GithubLatestVersionProvider(VersionProvider):
    @staticmethod
    def get_semantic_version(
        repo_name: str, tags: bool = False, request_session=set_up_retry(), io=ClickIOProvider()
    ) -> Union[SemanticVersion, None]:

        semantic_version = None
        try:
            if tags:
                response = request_session.get(f"https://api.github.com/repos/{repo_name}/tags")

                versions = [SemanticVersion.from_string(v["name"]) for v in response.json()]
                versions.sort(reverse=True)
                semantic_version = versions[0]
            else:
                package_info = request_session.get(
                    f"https://api.github.com/repos/{repo_name}/releases/latest"
                ).json()
                semantic_version = SemanticVersion.from_string(package_info["tag_name"])
        except Exception as e:
            io.error(f"Exception occured when calling Github with:\n{str(e)}")

        return semantic_version


class PyPiLatestVersionProvider(VersionProvider):
    @staticmethod
    def get_semantic_version(
        project_name: str, request_session=set_up_retry(), io=ClickIOProvider()
    ) -> Union[SemanticVersion, None]:
        semantic_version = None
        try:
            package_info = request_session.get(f"https://pypi.org/pypi/{project_name}/json").json()
            released_versions = package_info["releases"].keys()
            parsed_released_versions = [SemanticVersion.from_string(v) for v in released_versions]
            parsed_released_versions.sort(reverse=True)
            semantic_version = parsed_released_versions[0]
        except Exception as e:
            io.error(f"Exception occured when calling PyPi with:\n{str(e)}")
        return semantic_version


class AWSCLIInstalledVersionProvider(VersionProvider):
    @staticmethod
    def get_semantic_version() -> Union[SemanticVersion, None]:
        installed_aws_version = None
        try:
            response = subprocess.run("aws --version", capture_output=True, shell=True)
            matched = re.match(r"aws-cli/([0-9.]+)", response.stdout.decode("utf8"))
            installed_aws_version = matched.group(1)
        except (ValueError, AttributeError):
            pass
        return SemanticVersion.from_string(installed_aws_version)


class CopilotInstalledVersionProvider(VersionProvider):
    @staticmethod
    def get_semantic_version() -> Union[SemanticVersion, None]:
        copilot_version = None

        try:
            response = subprocess.run("copilot --version", capture_output=True, shell=True)
            [copilot_version] = re.findall(r"[0-9.]+", response.stdout.decode("utf8"))
        except ValueError:
            pass

        return SemanticVersion.from_string(copilot_version)



#!/usr/bin/env python
import click

from dbt_platform_helper.commands.copilot import make_addons
from dbt_platform_helper.commands.pipeline import generate as pipeline_generate
from dbt_platform_helper.domain.versioning import PlatformHelperVersioning
from dbt_platform_helper.utils.click import ClickDocOptCommand


@click.command(cls=ClickDocOptCommand)
@click.pass_context
def generate(ctx: click.Context):
    """
    Generate deployment pipeline configuration files and generate addons
    CloudFormation template files for each environment.

    Wraps pipeline generate and make-addons.
    """
    PlatformHelperVersioning().check_platform_helper_version_mismatch()
    ctx.invoke(pipeline_generate)
    ctx.invoke(make_addons)



import botocore
from boto3 import Session

from dbt_platform_helper.providers.aws.exceptions import ImageNotFoundException
from dbt_platform_helper.providers.aws.exceptions import RepositoryNotFoundException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.utils.application import Application
from dbt_platform_helper.utils.aws import get_aws_session_or_abort


class ECRProvider:
    def __init__(self, session: Session = None, click_io: ClickIOProvider = ClickIOProvider()):
        self.session = session
        self.click_io = click_io

    def get_ecr_repo_names(self) -> list[str]:
        out = []
        for page in self._get_client().get_paginator("describe_repositories").paginate():
            out.extend([repo["repositoryName"] for repo in page.get("repositories", {})])
        return out

    def get_image_details(
        self, application: Application, codebase: str, image_ref: str
    ) -> list[dict]:
        """Check if image exists in AWS ECR, and return a list of dictionaries
        containing image metadata."""

        repository = f"{application.name}/{codebase}"

        try:
            image_info = self._get_client().describe_images(
                repositoryName=repository,
                imageIds=[{"imageTag": image_ref}],
            )

            self._check_image_details_exists(image_info, image_ref)

            return image_info.get("imageDetails")
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "ImageNotFoundException":
                raise ImageNotFoundException(image_ref)
            if e.response["Error"]["Code"] == "RepositoryNotFoundException":
                raise RepositoryNotFoundException(repository)

    def find_commit_tag(self, image_details: list[dict], image_ref: str) -> str:
        """Loop through imageTags list to query for an image tag starting with
        'commit-', and return that value if found."""

        if image_ref.startswith("commit-"):
            return image_ref

        if image_details:
            for image in image_details:
                image_tags = image.get("imageTags", {})
                for tag in image_tags:
                    if tag.startswith("commit-"):
                        self.click_io.info(
                            f'INFO: The tag "{image_ref}" is not a unique, commit-specific tag. Deploying the corresponding commit tag "{tag}" instead.'
                        )
                        return tag
        self.click_io.warn(
            f'WARNING: The AWS ECR image "{image_ref}" has no associated commit tag so deploying "{image_ref}". Note this could result in images with unintended or incompatible changes being deployed if new ECS Tasks for your service.'
        )
        return image_ref

    @staticmethod
    def _check_image_details_exists(image_info: dict, image_ref: str):
        """Error handling for any unexpected scenario where AWS ECR returns a
        malformed response."""

        if "imageDetails" not in image_info:
            raise ImageNotFoundException(image_ref)

    def _get_client(self):
        if not self.session:
            self.session = get_aws_session_or_abort()
        return self.session.client("ecr")



from collections import Counter
from collections import OrderedDict
from copy import deepcopy
from typing import Protocol

from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.schema_migrations.schema_v0_to_v1_migration import (
    SchemaV0ToV1Migration,
)
from dbt_platform_helper.providers.version import InstalledVersionProvider


class InvalidMigrationConfigurationException(PlatformException):
    pass


class SchemaMigrationProtocol(Protocol):
    def from_version(self) -> int: ...

    def migrate(self, platform_config: dict) -> dict: ...


# TODO: Possibly get this programmatically?
ALL_MIGRATIONS = [SchemaV0ToV1Migration()]


class Migrator:
    def __init__(
        self,
        migrations: list[SchemaMigrationProtocol],
        installed_version_provider: InstalledVersionProvider = InstalledVersionProvider,
        io_provider: ClickIOProvider = ClickIOProvider(),
    ):
        self.migrations = sorted(migrations, key=lambda m: m.from_version())
        self.installed_version_provider = installed_version_provider
        self.io_provider = io_provider
        from_version_counts = Counter([migration.from_version() for migration in self.migrations])
        duplicate_from_versions = [count for count in from_version_counts.values() if count > 1]

        if duplicate_from_versions:
            raise InvalidMigrationConfigurationException(
                "`from_version` parameters must be unique amongst migrations"
            )

    def migrate(self, platform_config: dict) -> dict:
        out = OrderedDict(deepcopy(platform_config))
        if "schema_version" not in out:
            out["schema_version"] = 0

        if "default_versions" in out:
            out.move_to_end("default_versions", last=False)
        if "schema_version" in out:
            out.move_to_end("schema_version", last=False)
        if "application" in out:
            out.move_to_end("application", last=False)

        for migration in self.migrations:
            migration_can_be_applied = migration.from_version() == out["schema_version"]
            if migration_can_be_applied:
                out = migration.migrate(out)
                schema_version = out["schema_version"]
                self.io_provider.info(
                    f"Migrating from platform config schema version {schema_version} to version {schema_version + 1}"
                )
                out["schema_version"] += 1

        if "default_versions" not in out:
            out["default_versions"] = {}

        out["default_versions"]["platform-helper"] = str(
            self.installed_version_provider.get_semantic_version("dbt-platform-helper")
        )

        self.io_provider.info("\nMigration complete")

        return dict(out)



from copy import deepcopy


class SchemaV0ToV1Migration:
    def from_version(self) -> int:
        return 0

    def migrate(self, platform_config: dict) -> dict:
        migrated_config = deepcopy(platform_config)

        self._remove_terraform_platform_modules_default_version(migrated_config)
        self._remove_versions_from_env_config(migrated_config)
        self._remove_to_account_and_from_account_from_database_copy(migrated_config)
        self._remove_pipeline_platform_helper_override(migrated_config)

        return migrated_config

    def _remove_versions_from_env_config(self, migrated_config: dict) -> None:
        for env_name, env in migrated_config.get("environments", {}).items():
            if env and "versions" in env:
                del env["versions"]

    def _remove_terraform_platform_modules_default_version(self, migrated_config: dict) -> None:
        if "default_versions" in migrated_config:
            default_versions = migrated_config["default_versions"]
            if "terraform-platform-modules" in default_versions:
                del default_versions["terraform-platform-modules"]

    def _remove_to_account_and_from_account_from_database_copy(self, migrated_config: dict) -> None:
        for extension_name, extension in migrated_config.get("extensions", {}).items():
            if extension.get("type") == "postgres" and "database_copy" in extension:
                for database_copy_block in extension["database_copy"]:
                    if "from_account" in database_copy_block:
                        del database_copy_block["from_account"]
                    if "to_account" in database_copy_block:
                        del database_copy_block["to_account"]

    def _remove_pipeline_platform_helper_override(self, migrated_config: dict) -> None:
        for pipeline_name, pipeline_config in migrated_config.get(
            "environment_pipelines", {}
        ).items():
            if "versions" in pipeline_config:
                del pipeline_config["versions"]



from dbt_platform_helper.platform_exception import PlatformException


class AWSException(PlatformException):
    pass


class CreateTaskTimeoutException(AWSException):
    def __init__(self, addon_name: str, application_name: str, environment: str):
        super().__init__(
            f"""Client ({addon_name}) ECS task has failed to start for "{application_name}" in "{environment}" environment."""
        )


class ImageNotFoundException(AWSException):
    def __init__(self, image_ref: str):
        super().__init__(
            f"""An image labelled "{image_ref}" could not be found in your image repository. Try the `platform-helper codebase build` command first."""
        )


class RepositoryNotFoundException(AWSException):
    def __init__(self, repository: str):
        super().__init__(f"""The ECR repository "{repository}" could not be found.""")


class LogGroupNotFoundException(AWSException):
    def __init__(self, log_group_name: str):
        super().__init__(f"""No log group called "{log_group_name}".""")


# TODO: DBTP-1976: This should probably be in the AWS Copilot provider, but was causing circular import when we tried it pre refactoring the utils/aws.py
class CopilotCodebaseNotFoundException(PlatformException):
    def __init__(self, codebase: str):
        super().__init__(
            f"""The codebase "{codebase}" either does not exist or has not been deployed."""
        )


class CreateAccessTokenException(AWSException):
    def __init__(self, client_id: str):
        super().__init__(f"""Failed to create access token for Client "{client_id}".""")


class UnableToRetrieveSSOAccountList(AWSException):
    def __init__(self):
        super().__init__("Unable to retrieve AWS SSO account list")



import os
from collections.abc import Callable
from os import makedirs
from pathlib import Path
from shutil import rmtree

from dbt_platform_helper.constants import CODEBASE_PIPELINES_KEY
from dbt_platform_helper.constants import ENVIRONMENT_PIPELINES_KEY
from dbt_platform_helper.constants import PLATFORM_HELPER_VERSION_OVERRIDE_KEY
from dbt_platform_helper.constants import SUPPORTED_AWS_PROVIDER_VERSION
from dbt_platform_helper.constants import SUPPORTED_TERRAFORM_VERSION
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.ecr import ECRProvider
from dbt_platform_helper.providers.files import FileProvider
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.terraform_manifest import TerraformManifestProvider
from dbt_platform_helper.utils.application import get_application_name
from dbt_platform_helper.utils.template import setup_templates


class Pipelines:
    def __init__(
        self,
        config_provider: ConfigProvider,
        terraform_manifest_provider: TerraformManifestProvider,
        ecr_provider: ECRProvider,
        get_git_remote: Callable[[], str],
        get_codestar_arn: Callable[[str], str],
        io: ClickIOProvider = ClickIOProvider(),
        file_provider: FileProvider = FileProvider(),
        platform_helper_version_override: str = None,
    ):
        self.config_provider = config_provider
        self.get_git_remote = get_git_remote
        self.get_codestar_arn = get_codestar_arn
        self.terraform_manifest_provider = terraform_manifest_provider
        self.ecr_provider = ecr_provider
        self.io = io
        self.file_provider = file_provider
        self.platform_helper_version_override = platform_helper_version_override or os.environ.get(
            PLATFORM_HELPER_VERSION_OVERRIDE_KEY
        )

    def generate(
        self,
        deploy_branch: str,
    ):
        platform_config = self.config_provider.load_and_validate_platform_config()

        has_codebase_pipelines = CODEBASE_PIPELINES_KEY in platform_config
        has_environment_pipelines = ENVIRONMENT_PIPELINES_KEY in platform_config

        if not (has_codebase_pipelines or has_environment_pipelines):
            self.io.warn("No pipelines defined: nothing to do.")
            return

        app_name = get_application_name()

        git_repo = self.get_git_remote()
        if not git_repo:
            self.io.abort_with_error("The current directory is not a git repository")

        codestar_connection_arn = self.get_codestar_arn(app_name)
        if codestar_connection_arn is None:
            self.io.abort_with_error(f'There is no CodeStar Connection named "{app_name}" to use')

        base_path = Path(".")
        copilot_pipelines_dir = base_path / f"copilot/pipelines"

        self._clean_pipeline_config(copilot_pipelines_dir)

        platform_helper_version_for_template: str = platform_config.get("default_versions", {}).get(
            "platform-helper"
        )

        if self.platform_helper_version_override:
            platform_helper_version_for_template = self.platform_helper_version_override

        # TODO: DBTP-1965: - this whole code block/if-statement can fall away once the deploy_repository is a required key.
        deploy_repository = ""
        if "deploy_repository" in platform_config.keys():
            deploy_repository = f"{platform_config['deploy_repository']}"
        else:
            self.io.warn(
                "No `deploy_repository` key set in platform-config.yml, this will become a required key. See full platform config reference in the docs: https://platform.readme.trade.gov.uk/reference/platform-config-yml/#core-configuration"
            )
            deploy_repository = f"uktrade/{platform_config['application']}-deploy"

        if has_environment_pipelines:
            environment_pipelines = platform_config[ENVIRONMENT_PIPELINES_KEY]
            accounts = {
                config.get("account")
                for config in environment_pipelines.values()
                if "account" in config
            }

            for account in accounts:
                self._generate_terraform_environment_pipeline_manifest(
                    platform_config["application"],
                    deploy_repository,
                    account,
                    platform_helper_version_for_template,
                    deploy_branch,
                )

        if has_codebase_pipelines:
            codebase_pipelines = platform_config[CODEBASE_PIPELINES_KEY]
            ecrs_to_be_managed = {
                codebase: f"{platform_config['application']}/{codebase}"
                for codebase in codebase_pipelines.keys()
            }
            ecrs_already_provisioned = set(self.ecr_provider.get_ecr_repo_names())
            ecrs_that_need_importing = {
                codebase: repo
                for codebase, repo in ecrs_to_be_managed.items()
                if repo in ecrs_already_provisioned
            }

            self.terraform_manifest_provider.generate_codebase_pipeline_config(
                platform_config,
                platform_helper_version_for_template,
                ecrs_that_need_importing,
                deploy_repository,
            )

    def _clean_pipeline_config(self, pipelines_dir: Path):
        if pipelines_dir.exists():
            self.io.info("Deleting copilot/pipelines directory.")
            rmtree(pipelines_dir)

    def _generate_terraform_environment_pipeline_manifest(
        self,
        application: str,
        deploy_repository: str,
        aws_account: str,
        platform_helper_version: str,
        deploy_branch: str,
    ):
        env_pipeline_template = setup_templates().get_template("environment-pipelines/main.tf")

        contents = env_pipeline_template.render(
            {
                "application": application,
                "deploy_repository": deploy_repository,
                "aws_account": aws_account,
                "platform_helper_version": platform_helper_version,
                "deploy_branch": deploy_branch,
                "terraform_version": SUPPORTED_TERRAFORM_VERSION,
                "aws_provider_version": SUPPORTED_AWS_PROVIDER_VERSION,
            }
        )

        dir_path = f"terraform/environment-pipelines/{aws_account}"
        makedirs(dir_path, exist_ok=True)

        self.io.info(
            self.file_provider.mkfile(".", f"{dir_path}/main.tf", contents, overwrite=True)
        )



from dataclasses import dataclass

from dbt_platform_helper.platform_exception import PlatformException


class VpcProviderException(PlatformException):
    pass


class SubnetsNotFoundException(VpcProviderException):
    pass


class PrivateSubnetsNotFoundException(VpcProviderException):
    pass


class PublicSubnetsNotFoundException(VpcProviderException):
    pass


class SecurityGroupNotFoundException(VpcProviderException):
    pass


class VpcNotFoundForNameException(VpcProviderException):
    pass


class VpcIdMissingException(VpcProviderException):
    pass


@dataclass
class Vpc:
    id: str
    public_subnets: list[str]
    private_subnets: list[str]
    security_groups: list[str]


class VpcProvider:
    def __init__(self, session):
        self.ec2_client = session.client("ec2")

    def _get_subnet_ids(self, vpc_id):
        subnets = self.ec2_client.describe_subnets(
            Filters=[{"Name": "vpc-id", "Values": [vpc_id]}]
        ).get("Subnets")

        if not subnets:
            raise SubnetsNotFoundException(f"No subnets found for VPC with id: {vpc_id}.")

        public_tag = {"Key": "subnet_type", "Value": "public"}
        public_subnets = [subnet["SubnetId"] for subnet in subnets if public_tag in subnet["Tags"]]

        private_tag = {"Key": "subnet_type", "Value": "private"}
        private_subnets = [
            subnet["SubnetId"] for subnet in subnets if private_tag in subnet["Tags"]
        ]

        return public_subnets, private_subnets

    def _get_vpc_id_by_name(self, vpc_name: str) -> str:
        vpcs = self.ec2_client.describe_vpcs(
            Filters=[{"Name": "tag:Name", "Values": [vpc_name]}]
        ).get("Vpcs", [])

        if not vpcs:
            raise VpcNotFoundForNameException(f"VPC not found for name '{vpc_name}'")

        vpc_id = vpcs[0].get("VpcId")

        if not vpc_id:
            raise VpcIdMissingException(f"VPC id not present in vpc '{vpc_name}'")

        return vpc_id

    def _get_security_groups(self, app: str, env: str, vpc_id: str) -> list:

        vpc_filter = {"Name": "vpc-id", "Values": [vpc_id]}
        tag_filter = {"Name": f"tag:Name", "Values": [f"copilot-{app}-{env}-env"]}
        response = self.ec2_client.describe_security_groups(Filters=[vpc_filter, tag_filter])

        return [sg.get("GroupId") for sg in response.get("SecurityGroups")]

    def get_vpc(self, app: str, env: str, vpc_name: str) -> Vpc:

        vpc_id = self._get_vpc_id_by_name(vpc_name)

        public_subnets, private_subnets = self._get_subnet_ids(vpc_id)

        if not private_subnets:
            raise PrivateSubnetsNotFoundException(f"No private subnets found in vpc '{vpc_name}'")

        if not public_subnets:
            raise PublicSubnetsNotFoundException(f"No public subnets found in vpc '{vpc_name}'")

        sec_groups = self._get_security_groups(app, env, vpc_id)

        return Vpc(vpc_id, public_subnets, private_subnets, sec_groups)



#!/usr/bin/env python

import click

from dbt_platform_helper.domain.copilot import Copilot
from dbt_platform_helper.domain.copilot_environment import CopilotTemplating
from dbt_platform_helper.domain.versioning import PlatformHelperVersioning
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.config_validator import ConfigValidator
from dbt_platform_helper.providers.files import FileProvider
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.kms import KMSProvider
from dbt_platform_helper.providers.parameter_store import ParameterStore
from dbt_platform_helper.utils.aws import get_aws_session_or_abort
from dbt_platform_helper.utils.click import ClickDocOptGroup


@click.group(chain=True, cls=ClickDocOptGroup)
def copilot():
    PlatformHelperVersioning().check_if_needs_update()


@copilot.command()
def make_addons():
    """Generate addons CloudFormation for each environment."""
    try:
        session = get_aws_session_or_abort()
        parameter_provider = ParameterStore(session.client("ssm"))
        config_provider = ConfigProvider(ConfigValidator())
        Copilot(
            config_provider,
            parameter_provider,
            FileProvider(),
            CopilotTemplating(),
            KMSProvider,
            session,
        ).make_addons()
    except Exception as err:
        ClickIOProvider().abort_with_error(str(err))



from dbt_platform_helper.providers.slack_channel_notifier import SlackChannelNotifier
from dbt_platform_helper.utils.arn_parser import ARN


class Notify:
    def __init__(self, notifier: SlackChannelNotifier):
        self.notifier = notifier

    def environment_progress(
        self,
        message: str,
        build_arn: str = None,
        repository: str = None,
        commit_sha: str = None,
        original_message_ref: str = None,
    ):
        context = []

        if repository:
            context.append(f"*Repository*: <https://github.com/{repository}|{repository}>")
            if commit_sha:
                context.append(
                    f"*Revision*: <https://github.com/{repository}/commit/{commit_sha}|{commit_sha}>"
                )

        if build_arn:
            context.append(f"<{get_build_url(build_arn)}|Build Logs>")

        if original_message_ref:
            return self.notifier.post_update(original_message_ref, message, context)
        else:
            return self.notifier.post_new(message, context)

    def add_comment(
        self,
        original_message_ref: str,
        message: str,
        title: str,
        reply_broadcast: bool,
    ):
        self.notifier.post_new(
            message=message,
            title=title,
            context=[],
            reply_broadcast=reply_broadcast,
            thread_ref=original_message_ref,
        )


def get_build_url(build_arn: str):
    try:
        arn = ARN(build_arn)
        url = (
            "https://{region}.console.aws.amazon.com/codesuite/codebuild/{account}/projects/{"
            "project}/build/{project}%3A{build_id}"
        )
        return url.format(
            region=arn.region,
            account=arn.account_id,
            project=arn.project.replace("build/", ""),
            build_id=arn.build_id,
        )
    except ValueError:
        return ""



import boto3

from dbt_platform_helper.platform_exception import PlatformException


class ParameterStore:
    def __init__(self, ssm_client: boto3.client):
        self.ssm_client = ssm_client

    def get_ssm_parameter_by_name(self, parameter_name: str) -> dict:
        """
        Retrieves the latest version of a parameter from parameter store for a
        given name/arn.

        Args:
            path (str): The parameter name to retrieve the parameter value for.
        Returns:
            dict: A dictionary representation of your ssm parameter
        """

        return self.ssm_client.get_parameter(Name=parameter_name)["Parameter"]

    def get_ssm_parameters_by_path(self, path: str) -> list:
        """
        Retrieves all SSM parameters for a given path from parameter store.

        Args:
            path (str): The parameter path to retrieve the parameters for. e.g. /copilot/applications/
        Returns:
            list: A list of dictionaries containing all SSM parameters under the provided path.
        """

        parameters = []
        paginator = self.ssm_client.get_paginator("get_parameters_by_path")
        page_iterator = paginator.paginate(Path=path, Recursive=True)

        for page in page_iterator:
            parameters.extend(page.get("Parameters", []))

        if parameters:
            return parameters
        else:
            raise ParameterNotFoundForPathException()


class ParameterNotFoundForPathException(PlatformException):
    """Exception raised when no parameters are found for a given path."""



import random
import string
import time
from typing import List

from dbt_platform_helper.platform_exception import PlatformException


class ECS:
    def __init__(self, ecs_client, ssm_client, application_name: str, env: str):
        self.ecs_client = ecs_client
        self.ssm_client = ssm_client
        self.application_name = application_name
        self.env = env

    def get_cluster_arn(self) -> str:
        """Returns the ARN of the ECS cluster for the given application and
        environment."""
        for cluster_arn in self.ecs_client.list_clusters()["clusterArns"]:
            tags_response = self.ecs_client.list_tags_for_resource(resourceArn=cluster_arn)
            tags = tags_response["tags"]

            app_key_found = False
            env_key_found = False
            cluster_key_found = False

            for tag in tags:
                if tag["key"] == "copilot-application" and tag["value"] == self.application_name:
                    app_key_found = True
                if tag["key"] == "copilot-environment" and tag["value"] == self.env:
                    env_key_found = True
                if tag["key"] == "aws:cloudformation:logical-id" and tag["value"] == "Cluster":
                    cluster_key_found = True

            if app_key_found and env_key_found and cluster_key_found:
                return cluster_arn

        raise NoClusterException(self.application_name, self.env)

    def get_or_create_task_name(self, addon_name: str, parameter_name: str) -> str:
        """Fetches the task name from SSM or creates a new one if not found."""
        try:
            return self.ssm_client.get_parameter(Name=parameter_name)["Parameter"]["Value"]
        except self.ssm_client.exceptions.ParameterNotFound:
            random_id = "".join(random.choices(string.ascii_lowercase + string.digits, k=12))
            return f"conduit-{self.application_name}-{self.env}-{addon_name}-{random_id}"

    def get_ecs_task_arns(self, cluster_arn: str, task_name: str):
        """Gets the ECS task ARNs for a given task name and cluster ARN."""
        tasks = self.ecs_client.list_tasks(
            cluster=cluster_arn,
            desiredStatus="RUNNING",
            family=f"copilot-{task_name}",
        )

        if not tasks["taskArns"]:
            return []

        return tasks["taskArns"]

    def ecs_exec_is_available(self, cluster_arn: str, task_arns: List[str]):
        """
        Checks if the ExecuteCommandAgent is running on the specified ECS task.

        Waits for up to 25 attempts, then raises ECSAgentNotRunning if still not
        running.
        """
        current_attempts = 0
        execute_command_agent_status = ""

        while execute_command_agent_status != "RUNNING" and current_attempts < 25:
            current_attempts += 1

            task_details = self.ecs_client.describe_tasks(cluster=cluster_arn, tasks=task_arns)

            managed_agents = task_details["tasks"][0]["containers"][0]["managedAgents"]
            execute_command_agent_status = [
                agent["lastStatus"]
                for agent in managed_agents
                if agent["name"] == "ExecuteCommandAgent"
            ][0]
            if execute_command_agent_status != "RUNNING":
                time.sleep(1)

        if execute_command_agent_status != "RUNNING":
            raise ECSAgentNotRunningException


class ECSException(PlatformException):
    pass


class ECSAgentNotRunningException(ECSException):
    def __init__(self):
        super().__init__("""ECS exec agent never reached "RUNNING" status""")


class NoClusterException(ECSException):
    def __init__(self, application_name: str, environment: str):
        super().__init__(
            f"""No ECS cluster found for "{application_name}" in "{environment}" environment."""
        )



import click

from dbt_platform_helper.domain.conduit import Conduit
from dbt_platform_helper.domain.versioning import PlatformHelperVersioning
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.cloudformation import CloudFormation
from dbt_platform_helper.providers.ecs import ECS
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.secrets import Secrets
from dbt_platform_helper.utils.application import load_application
from dbt_platform_helper.utils.click import ClickDocOptCommand

CONDUIT_ACCESS_OPTIONS = ["read", "write", "admin"]


@click.command(cls=ClickDocOptCommand)
@click.argument("addon_name", type=str, required=True)
@click.option("--app", help="Application name", required=True)
@click.option("--env", help="Environment name", required=True)
@click.option(
    "--access",
    default="read",
    type=click.Choice(CONDUIT_ACCESS_OPTIONS),
    help="Allow read, write or admin access to the database addons.",
)
def conduit(addon_name: str, app: str, env: str, access: str):
    """Opens a shell for a given addon_name create a conduit connection to
    interact with postgres, opensearch or redis."""
    PlatformHelperVersioning().check_if_needs_update()
    application = load_application(app)

    try:
        secrets_provider: Secrets = Secrets(
            application.environments[env].session.client("ssm"),
            application.environments[env].session.client("secretsmanager"),
            application.name,
            env,
        )
        cloudformation_provider: CloudFormation = CloudFormation(
            application.environments[env].session.client("cloudformation"),
            application.environments[env].session.client("iam"),
            application.environments[env].session.client("ssm"),
        )

        ecs_provider: ECS = ECS(
            application.environments[env].session.client("ecs"),
            application.environments[env].session.client("ssm"),
            application.name,
            env,
        )

        Conduit(application, secrets_provider, cloudformation_provider, ecs_provider).start(
            env, addon_name, access
        )
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))



from boto3 import Session

from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.utils.aws import get_aws_session_or_abort


def normalise_to_cidr(ip: str):
    if "/" in ip:
        return ip
    SINGLE_IPV4_CIDR_PREFIX_LENGTH = "32"
    return f"{ip}/{SINGLE_IPV4_CIDR_PREFIX_LENGTH}"


class LoadBalancerProvider:

    def __init__(self, session: Session = None, io: ClickIOProvider = ClickIOProvider()):
        self.session = session
        self.evlb_client = self._get_client("elbv2")
        self.rg_tagging_client = self._get_client("resourcegroupstaggingapi")
        self.io = io

    def _get_client(self, client: str):
        if not self.session:
            self.session = get_aws_session_or_abort()
        return self.session.client(client)

    def find_target_group(self, app: str, env: str, svc: str) -> str:
        target_group_arn = None

        response = self.rg_tagging_client.get_resources(
            TagFilters=[
                {
                    "Key": "copilot-application",
                    "Values": [
                        app,
                    ],
                    "Key": "copilot-environment",
                    "Values": [
                        env,
                    ],
                    "Key": "copilot-service",
                    "Values": [
                        svc,
                    ],
                },
            ],
            ResourceTypeFilters=[
                "elasticloadbalancing:targetgroup",
            ],
        )  # TODO: DBTP-1942: should be paginated
        for resource in response["ResourceTagMappingList"]:
            tags = {tag["Key"]: tag["Value"] for tag in resource["Tags"]}

            if (
                "copilot-service" in tags
                and tags["copilot-service"] == svc
                and "copilot-environment" in tags
                and tags["copilot-environment"] == env
                and "copilot-application" in tags
                and tags["copilot-application"] == app
            ):
                target_group_arn = resource["ResourceARN"]

        if not target_group_arn:
            self.io.error(
                f"No target group found for application: {app}, environment: {env}, service: {svc}",
            )

        return target_group_arn

    def get_https_certificate_for_application(self, app: str, env: str) -> str:
        listener_arn = self.get_https_listener_for_application(app, env)
        certificates = self.evlb_client.describe_listener_certificates(ListenerArn=listener_arn)[
            "Certificates"
        ]  # TODO: DBTP-1942: should be paginated

        try:
            certificate_arn = next(c["CertificateArn"] for c in certificates if c["IsDefault"])
        except StopIteration:
            raise CertificateNotFoundException(env)

        return certificate_arn

    def get_https_listener_for_application(self, app: str, env: str) -> str:
        load_balancer_arn = self.get_load_balancer_for_application(app, env)

        listeners = self.evlb_client.describe_listeners(LoadBalancerArn=load_balancer_arn)[
            "Listeners"
        ]  # TODO: DBTP-1942: should be paginated

        listener_arn = None

        try:
            listener_arn = next(l["ListenerArn"] for l in listeners if l["Protocol"] == "HTTPS")
        except StopIteration:
            pass

        if not listener_arn:
            raise ListenerNotFoundException(app, env)

        return listener_arn

    def get_load_balancer_for_application(self, app: str, env: str) -> str:
        describe_response = self.evlb_client.describe_load_balancers()
        load_balancers = [lb["LoadBalancerArn"] for lb in describe_response["LoadBalancers"]]

        tag_descriptions = []
        for i in range(0, len(load_balancers), 20):
            chunk = load_balancers[i : i + 20]
            tag_descriptions.extend(
                self.evlb_client.describe_tags(ResourceArns=chunk)["TagDescriptions"]
            )

        for lb in tag_descriptions:
            tags = {t["Key"]: t["Value"] for t in lb["Tags"]}
            # TODO: DBTP-1967: copilot hangover, creates coupling to specific tags could update to check application and environment
            if tags.get("copilot-application") == app and tags.get("copilot-environment") == env:
                return lb["ResourceArn"]

        raise LoadBalancerNotFoundException(app, env)

    def get_host_header_conditions(self, listener_arn: str, target_group_arn: str) -> list:
        rules = self.evlb_client.describe_rules(ListenerArn=listener_arn)[
            "Rules"
        ]  # TODO: DBTP-1942: should be paginated

        conditions = []

        for rule in rules:
            for action in rule["Actions"]:
                if action["Type"] == "forward" and action["TargetGroupArn"] == target_group_arn:
                    conditions = rule["Conditions"]

        if not conditions:
            raise ListenerRuleConditionsNotFoundException(listener_arn)

        # filter to host-header conditions
        conditions = [
            {i: condition[i] for i in condition if i != "Values"}
            for condition in conditions
            if condition["Field"] == "host-header"
        ]

        # remove internal hosts
        conditions[0]["HostHeaderConfig"]["Values"] = [
            v for v in conditions[0]["HostHeaderConfig"]["Values"]
        ]

        return conditions

    def get_rules_tag_descriptions_by_listener_arn(self, listener_arn: str) -> list:
        rules = self.evlb_client.describe_rules(ListenerArn=listener_arn)[
            "Rules"
        ]  # TODO: DBTP-1942: should be paginated
        return self.get_rules_tag_descriptions(rules)

    def get_rules_tag_descriptions(self, rules: list) -> list:
        tag_descriptions = []
        chunk_size = 20

        for i in range(0, len(rules), chunk_size):
            chunk = rules[i : i + chunk_size]
            resource_arns = [r["RuleArn"] for r in chunk]
            response = self.evlb_client.describe_tags(
                ResourceArns=resource_arns
            )  # TODO should be paginated
            tag_descriptions.extend(response["TagDescriptions"])

        return tag_descriptions

    def create_rule(
        self,
        listener_arn: str,
        actions: list,
        conditions: list,
        priority: int,
        tags: list,
    ):
        return self.evlb_client.create_rule(
            ListenerArn=listener_arn,
            Priority=priority,
            Conditions=conditions,
            Actions=actions,
            Tags=tags,
        )

    def create_forward_rule(
        self,
        listener_arn: str,
        target_group_arn: str,
        rule_name: str,
        priority: int,
        conditions: list,
        additional_tags: list = [],
    ):
        return self.create_rule(
            listener_arn=listener_arn,
            priority=priority,
            conditions=conditions,
            actions=[{"Type": "forward", "TargetGroupArn": target_group_arn}],
            tags=[{"Key": "name", "Value": rule_name}, *additional_tags],
        )

    def create_header_rule(
        self,
        listener_arn: str,
        target_group_arn: str,
        header_name: str,
        values: list,
        rule_name: str,
        priority: int,
        conditions: list,
        additional_tags: list = [],
    ):

        combined_conditions = [
            {
                "Field": "http-header",
                "HttpHeaderConfig": {"HttpHeaderName": header_name, "Values": values},
            }
        ] + conditions

        self.create_forward_rule(
            listener_arn,
            target_group_arn,
            rule_name,
            priority,
            combined_conditions,
            additional_tags,
        )

        self.io.debug(
            f"Creating listener rule {rule_name} for HTTPS Listener with arn {listener_arn}.\nIf request header {header_name} contains one of the values {values}, the request will be forwarded to target group with arn {target_group_arn}.\n\n",
        )

    def create_source_ip_rule(
        self,
        listener_arn: str,
        target_group_arn: str,
        values: list,
        rule_name: str,
        priority: int,
        conditions: list,
        additional_tags: list = [],
    ):
        combined_conditions = [
            {
                "Field": "source-ip",
                "SourceIpConfig": {"Values": [normalise_to_cidr(value) for value in values]},
            }
        ] + conditions

        self.create_forward_rule(
            listener_arn,
            target_group_arn,
            rule_name,
            priority,
            combined_conditions,
            additional_tags,
        )

        self.io.debug(
            f"Creating listener rule {rule_name} for HTTPS Listener with arn {listener_arn}.\nIf request source ip matches one of the values {values}, the request will be forwarded to target group with arn {target_group_arn}.\n\n",
        )

    def delete_listener_rule_by_tags(self, tag_descriptions: list, tag_name: str) -> list:
        deleted_rules = []

        for description in tag_descriptions:
            tags = {t["Key"]: t["Value"] for t in description["Tags"]}
            if tags.get("name") == tag_name:
                if description["ResourceArn"]:
                    self.evlb_client.delete_rule(RuleArn=description["ResourceArn"])
                    deleted_rules.append(description)

        return deleted_rules


class LoadBalancerException(PlatformException):
    pass


class LoadBalancerNotFoundException(LoadBalancerException):
    def __init__(self, application_name, env):
        super().__init__(
            f"No load balancer found for environment {env} in the application {application_name}."
        )


class ListenerNotFoundException(LoadBalancerException):
    def __init__(self, application_name, env):
        super().__init__(
            f"No HTTPS listener found for environment {env} in the application {application_name}."
        )


class ListenerRuleNotFoundException(LoadBalancerException):
    pass


class ListenerRuleConditionsNotFoundException(LoadBalancerException):
    def __init__(self, listener_arn):
        super().__init__(f"No listener rule conditions found for listener ARN: {listener_arn}")


class CertificateNotFoundException(PlatformException):
    def __init__(self, environment_name: str):
        super().__init__(
            f"""No certificate found with domain name matching environment {environment_name}."."""
        )



import re

import jinja2

from dbt_platform_helper.jinja2_tags import ExtraHeaderTag
from dbt_platform_helper.jinja2_tags import VersionTag

S3_CROSS_ACCOUNT_POLICY = "addons/svc/s3-cross-account-policy.yml"

ADDON_TEMPLATE_MAP = {
    "s3": ["addons/svc/s3-policy.yml"],
    "s3-policy": ["addons/svc/s3-policy.yml"],
    "appconfig-ipfilter": ["addons/svc/appconfig-ipfilter.yml"],
    "subscription-filter": ["addons/svc/subscription-filter.yml"],
    "prometheus-policy": ["addons/svc/prometheus-policy.yml"],
}


def camel_case(s):
    s = re.sub(r"(_|-)+", " ", s).title().replace(" ", "")
    return "".join([s[0].lower(), s[1:]])


def setup_templates() -> jinja2.Environment:
    templateLoader = jinja2.PackageLoader("dbt_platform_helper")
    templateEnv = jinja2.Environment(loader=templateLoader, keep_trailing_newline=True)
    templateEnv.add_extension(ExtraHeaderTag)
    templateEnv.add_extension(VersionTag)

    return templateEnv



import os

from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.semantic_version import (
    IncompatibleMajorVersionException,
)
from dbt_platform_helper.providers.semantic_version import (
    IncompatibleMinorVersionException,
)
from dbt_platform_helper.providers.semantic_version import SemanticVersion
from dbt_platform_helper.providers.version import AWSCLIInstalledVersionProvider
from dbt_platform_helper.providers.version import CopilotInstalledVersionProvider
from dbt_platform_helper.providers.version import GithubLatestVersionProvider
from dbt_platform_helper.providers.version import InstalledVersionProvider
from dbt_platform_helper.providers.version import PyPiLatestVersionProvider
from dbt_platform_helper.providers.version import VersionProvider
from dbt_platform_helper.providers.version_status import VersionStatus


def running_as_installed_package():
    return "site-packages" in __file__


def skip_version_checks():
    return not running_as_installed_package() or "PLATFORM_TOOLS_SKIP_VERSION_CHECK" in os.environ


class PlatformHelperVersionNotFoundException(PlatformException):
    def __init__(self, message=None):
        super().__init__(message or "Platform helper version could not be resolved.")


class PlatformHelperVersioning:
    def __init__(
        self,
        io: ClickIOProvider = ClickIOProvider(),
        config_provider: ConfigProvider = ConfigProvider(),
        latest_version_provider: VersionProvider = PyPiLatestVersionProvider,
        installed_version_provider: InstalledVersionProvider = InstalledVersionProvider(),
        skip_versioning_checks: bool = None,
    ):
        self.io = io
        self.config_provider = config_provider
        self.latest_version_provider = latest_version_provider
        self.installed_version_provider = installed_version_provider
        self.skip_versioning_checks = (
            skip_versioning_checks if skip_versioning_checks is not None else skip_version_checks()
        )

    def get_required_version(self):
        platform_config = self.config_provider.load_and_validate_platform_config()
        required_version = platform_config.get("default_versions", {}).get("platform-helper")
        self.io.info(required_version)
        return required_version

    # Used in the generate command
    def check_platform_helper_version_mismatch(self):
        if self.skip_versioning_checks:
            return

        version_status = self.get_version_status()
        required_version = self.get_required_version()

        if SemanticVersion.is_semantic_version(required_version):
            required_version_semver = SemanticVersion.from_string(required_version)

            if not version_status.installed == required_version_semver:
                message = (
                    f"WARNING: You are running platform-helper v{version_status.installed} against "
                    f"v{required_version_semver} specified for the project."
                )
                self.io.warn(message)

    def check_if_needs_update(self):
        if self.skip_versioning_checks:
            return

        version_status = self.get_version_status()

        message = (
            f"You are running platform-helper v{version_status.installed}, upgrade to "
            f"v{version_status.latest} by running run `pip install "
            "--upgrade dbt-platform-helper`."
        )

        try:
            version_status.installed.validate_compatibility_with(version_status.latest)
        except IncompatibleMajorVersionException:
            self.io.error(message)
        except IncompatibleMinorVersionException:
            self.io.warn(message)

    def get_version_status(self) -> VersionStatus:
        locally_installed_version = self.installed_version_provider.get_semantic_version(
            "dbt-platform-helper"
        )

        latest_release = self.latest_version_provider.get_semantic_version("dbt-platform-helper")

        return VersionStatus(installed=locally_installed_version, latest=latest_release)


class AWSVersioning:
    def __init__(
        self,
        latest_version_provider: VersionProvider = None,
        installed_version_provider: VersionProvider = None,
    ):
        self.latest_version_provider = latest_version_provider or GithubLatestVersionProvider
        self.installed_version_provider = (
            installed_version_provider or AWSCLIInstalledVersionProvider
        )

    def get_version_status(self) -> VersionStatus:
        return VersionStatus(
            self.installed_version_provider.get_semantic_version(),
            self.latest_version_provider.get_semantic_version("aws/aws-cli", True),
        )


class CopilotVersioning:
    def __init__(
        self,
        latest_version_provider: VersionProvider = None,
        installed_version_provider: VersionProvider = None,
    ):
        self.latest_version_provider = latest_version_provider or GithubLatestVersionProvider
        self.installed_version_provider = (
            installed_version_provider or CopilotInstalledVersionProvider
        )

    def get_version_status(self) -> VersionStatus:
        return VersionStatus(
            self.installed_version_provider.get_semantic_version(),
            self.latest_version_provider.get_semantic_version("aws/copilot-cli"),
        )



import json

import botocore
from cfn_tools import dump_yaml
from cfn_tools import load_yaml

from dbt_platform_helper.platform_exception import PlatformException


class CloudFormation:
    # TODO: DBTP-1966: add handling for optional client parameters to handle case of calling boto API with None
    def __init__(self, cloudformation_client, iam_client=None, ssm_client=None):
        self.cloudformation_client = cloudformation_client
        self.iam_client = iam_client
        self.ssm_client = ssm_client

    def add_stack_delete_policy_to_task_role(self, task_name: str):
        stack_name = f"task-{task_name}"
        stack_resources = self.cloudformation_client.list_stack_resources(StackName=stack_name)[
            "StackResourceSummaries"
        ]

        for resource in stack_resources:
            if resource["LogicalResourceId"] == "DefaultTaskRole":
                task_role_name = resource["PhysicalResourceId"]
                self.iam_client.put_role_policy(
                    RoleName=task_role_name,
                    PolicyName="DeleteCloudFormationStack",
                    PolicyDocument=json.dumps(
                        {
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Action": ["cloudformation:DeleteStack"],
                                    "Effect": "Allow",
                                    "Resource": f"arn:aws:cloudformation:*:*:stack/{stack_name}/*",
                                },
                            ],
                        },
                    ),
                )

    def update_conduit_stack_resources(
        self,
        application_name: str,
        env: str,
        addon_type: str,
        addon_name: str,
        task_name: str,
        parameter_name: str,
        access: str,
    ):
        conduit_stack_name = f"task-{task_name}"
        template = self.cloudformation_client.get_template(StackName=conduit_stack_name)
        template_yml = load_yaml(template["TemplateBody"])

        template_yml["Resources"]["LogGroup"]["DeletionPolicy"] = "Retain"

        template_yml["Resources"]["TaskNameParameter"] = load_yaml(
            f"""
            Type: AWS::SSM::Parameter
            Properties:
              Name: {parameter_name}
              Type: String
              Value: {task_name}
            """
        )

        log_filter_role_arn = self.iam_client.get_role(RoleName="CWLtoSubscriptionFilterRole")[
            "Role"
        ]["Arn"]

        destination_log_group_arns = json.loads(
            self.ssm_client.get_parameter(Name="/copilot/tools/central_log_groups")["Parameter"][
                "Value"
            ]
        )

        destination_arn = destination_log_group_arns["dev"]
        if env.lower() in ("prod", "production"):
            destination_arn = destination_log_group_arns["prod"]

        template_yml["Resources"]["SubscriptionFilter"] = load_yaml(
            f"""
            Type: AWS::Logs::SubscriptionFilter
            DeletionPolicy: Retain
            Properties:
              RoleArn: {log_filter_role_arn}
              LogGroupName: /copilot/{task_name}
              FilterName: /copilot/conduit/{application_name}/{env}/{addon_type}/{addon_name}/{task_name.rsplit("-", 1)[1]}/{access}
              FilterPattern: ''
              DestinationArn: {destination_arn}
            """
        )

        params = []
        if "Parameters" in template_yml:
            for param in template_yml["Parameters"]:
                params.append({"ParameterKey": param, "UsePreviousValue": True})

        self.cloudformation_client.update_stack(
            StackName=conduit_stack_name,
            TemplateBody=dump_yaml(template_yml),
            Parameters=params,
            Capabilities=["CAPABILITY_IAM"],
        )

        return conduit_stack_name

    def wait_for_cloudformation_to_reach_status(self, stack_status, stack_name):
        waiter = self.cloudformation_client.get_waiter(stack_status)

        try:
            waiter.wait(StackName=stack_name, WaiterConfig={"Delay": 5, "MaxAttempts": 20})
        except botocore.exceptions.WaiterError as err:
            current_status = err.last_response.get("Stacks", [{}])[0].get("StackStatus", "")

            if current_status in [
                "ROLLBACK_IN_PROGRESS",
                "UPDATE_ROLLBACK_IN_PROGRESS",
                "ROLLBACK_FAILED",
            ]:
                raise CloudFormationException(stack_name, current_status)
            else:
                raise CloudFormationException(
                    stack_name, f"Error while waiting for stack status: {str(err)}"
                )

    def get_cloudformation_exports_for_environment(self, environment_name):
        exports = []

        for page in self.cloudformation_client.get_paginator("list_exports").paginate():
            for export in page["Exports"]:
                if f"-{environment_name}-" in export["Name"]:
                    exports.append(export)

        return exports


class CloudFormationException(PlatformException):
    def __init__(self, stack_name: str, current_status: str):
        super().__init__(
            f"The CloudFormation stack '{stack_name}' is not in a good state: {current_status}"
        )



#!/usr/bin/env python

import copy
import json
from pathlib import Path
from pathlib import PosixPath

import botocore
import botocore.errorfactory

from dbt_platform_helper.constants import PLATFORM_CONFIG_FILE
from dbt_platform_helper.domain.copilot_environment import CopilotTemplating
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.files import FileProvider
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.kms import KMSProvider
from dbt_platform_helper.providers.parameter_store import ParameterStore
from dbt_platform_helper.providers.yaml_file import YamlFileProvider
from dbt_platform_helper.utils.application import get_application_name
from dbt_platform_helper.utils.application import load_application
from dbt_platform_helper.utils.template import ADDON_TEMPLATE_MAP
from dbt_platform_helper.utils.template import camel_case
from dbt_platform_helper.utils.template import setup_templates
from dbt_platform_helper.utils.validation import validate_addons


class Copilot:

    PACKAGE_DIR = Path(__file__).resolve().parent.parent

    SERVICE_TYPES = [
        "Load Balanced Web Service",
        "Backend Service",
        "Request-Driven Web Service",
        "Static Site",
        "Worker Service",
    ]

    def __init__(
        self,
        config_provider: ConfigProvider,
        parameter_provider: ParameterStore,
        file_provider: FileProvider,
        copilot_templating: CopilotTemplating,
        kms_provider: KMSProvider,
        session,
        io: ClickIOProvider = ClickIOProvider(),
        yaml_file_provider: YamlFileProvider = YamlFileProvider,
    ):
        self.config_provider = config_provider
        self.parameter_provider = parameter_provider
        self.file_provider = file_provider
        self.copilot_templating = copilot_templating
        self.kms_provider = kms_provider
        self.io = io
        self.yaml_file_provider = yaml_file_provider
        self.session = session

    def make_addons(self):
        config = self.config_provider.load_and_validate_platform_config()

        templates = setup_templates()
        extensions = self._get_extensions()
        application_name = get_application_name()

        self.io.info("\n>>> Generating Terraform compatible addons CloudFormation\n")

        output_dir = Path(".").absolute()
        env_path = Path(f"copilot/environments/")
        env_addons_path = env_path / "addons"
        env_overrides_path = env_path / "overrides"

        self._cleanup_old_files(extensions, output_dir, env_addons_path, env_overrides_path)
        self._generate_env_overrides(output_dir)

        svc_names = self._list_copilot_local_services()
        base_path = Path(".")
        for svc_name in svc_names:
            self._generate_svc_overrides(base_path, templates, svc_name)

        services = []
        for ext_name, ext_data in extensions.items():
            extension = {**ext_data}
            addon_type = extension.pop("type")
            environments = extension.pop("environments")
            environment_addon_config = {
                "addon_type": addon_type,
                "environments": environments,
                "name": extension.get("name", None) or ext_name,
                "prefix": camel_case(ext_name),
                "secret_name": ext_name.upper().replace("-", "_"),
                **extension,
            }

            services.append(environment_addon_config)

            service_addon_config = {
                "application_name": application_name,
                "name": extension.get("name", None) or ext_name,
                "prefix": camel_case(ext_name),
                "environments": environments,
                **extension,
            }

            log_destination_arns = self._get_log_destination_arn()

            if addon_type in ["s3", "s3-policy"]:
                if extensions[ext_name].get("serve_static_content"):
                    continue

                s3_kms_arns = self._get_s3_kms_alias_arns(application_name, environments)
                for environment_name in environments:
                    environments[environment_name]["kms_key_arn"] = s3_kms_arns.get(
                        environment_name, "kms-key-not-found"
                    )

            self._generate_service_addons(
                extension,
                ext_name,
                addon_type,
                output_dir,
                service_addon_config,
                templates,
                log_destination_arns,
            )

        environments = self.config_provider.apply_environment_defaults(config)["environments"]

        self.copilot_templating.generate_cross_account_s3_policies(environments, extensions)

        self.io.info(templates.get_template("addon-instructions.txt").render(services=services))

    def _list_copilot_local_environments(self):
        return [
            path.parent.parts[-1] for path in Path("./copilot/environments/").glob("*/manifest.yml")
        ]

    def _is_service(self, path: PosixPath) -> bool:

        manifest_file = self.yaml_file_provider.load(path)
        if not manifest_file or not manifest_file.get("type"):
            self.io.abort_with_error(f"No type defined in manifest file {str(path)}; exiting")

        return manifest_file.get("type") in self.SERVICE_TYPES

    def _list_copilot_local_services(self):
        return [
            path.parent.parts[-1]
            for path in Path("./copilot/").glob("*/manifest.yml")
            if self._is_service(path)
        ]

    def _validate_and_normalise_extensions_config(self, config_file, key_in_config_file=None):
        """Load a config file, validate it against the extensions schemas and
        return the normalised config dict."""

        def _lookup_plan(addon_type, env_conf):
            plan = env_conf.pop("plan", None)
            conf = addon_plans[addon_type][plan] if plan else {}

            # Make a copy of the addon plan config so subsequent
            # calls do not override the root object
            conf = conf.copy()

            conf.update(env_conf)

            return conf

        def _normalise_keys(source: dict):
            return {k.replace("-", "_"): v for k, v in source.items()}

        addon_plans = self.yaml_file_provider.load(self.PACKAGE_DIR / "addon-plans.yml")

        # load and validate config
        config = self.yaml_file_provider.load(config_file)

        if config and key_in_config_file:
            config = config[key_in_config_file]

        # empty file
        if not config:
            return {}

        errors = validate_addons(config)

        if errors:
            self.io.error(f"Errors found in {config_file}:")
            for addon, error in errors.items():
                self.io.error(f"Addon '{addon}': {error}")
            self.io.abort_with_error("Invalid platform-config.yml provided, see above warnings")

        env_names = self._list_copilot_local_environments()
        svc_names = self._list_copilot_local_services()

        if not env_names:
            self.io.abort_with_error("No environments found in ./copilot/environments; exiting")

        if not svc_names:
            self.io.abort_with_error("No services found in ./copilot/; exiting")

        normalised_config = {}
        config_has_errors = False
        for addon_name, addon_config in config.items():
            addon_type = addon_config["type"]
            normalised_config[addon_name] = copy.deepcopy(addon_config)

            if "services" in normalised_config[addon_name]:
                if normalised_config[addon_name]["services"] == "__all__":
                    normalised_config[addon_name]["services"] = svc_names

                if not set(normalised_config[addon_name]["services"]).issubset(set(svc_names)):
                    self.io.error(
                        f"Services listed in {addon_name}.services do not exist in ./copilot/"
                    )
                    config_has_errors = True

            environments = normalised_config[addon_name].pop("environments", {})
            default = environments.pop("*", environments.pop("default", {}))

            initial = _lookup_plan(addon_type, default)

            missing_envs = set(environments.keys()) - set(env_names)
            if missing_envs:
                self.io.error(
                    f"Environment keys listed in {addon_name} do not match those defined in ./copilot/environments"
                )
                self.io.error(f"  Missing environments: {', '.join(sorted(missing_envs))}")
                config_has_errors = True

            if config_has_errors:
                continue

            normalised_environments = {}

            for env in env_names:
                normalised_environments[env] = _normalise_keys(initial)

            for env_name, env_config in environments.items():
                if env_config is None:
                    env_config = {}
                normalised_environments[env_name].update(
                    _lookup_plan(addon_type, _normalise_keys(env_config))
                )

            normalised_config[addon_name]["environments"] = normalised_environments

        if config_has_errors:
            self.io.abort_with_error("Configuration has errors. Exiting.")

        return normalised_config

    def _get_log_destination_arn(self):
        """Get destination arns stored in param store in projects aws
        account."""

        try:
            destination_arns = self.parameter_provider.get_ssm_parameter_by_name(
                "/copilot/tools/central_log_groups"
            )
        except botocore.errorfactory.ParameterNotFound:
            self.io.abort_with_error(
                "No aws central log group defined in Parameter Store at location /copilot/tools/central_log_groups; exiting"
            )

        return json.loads(destination_arns["Value"])

    def _generate_svc_overrides(self, base_path, templates, name):
        self.io.info(f"\n>>> Generating service overrides for {name}\n")
        overrides_path = base_path.joinpath(f"copilot/{name}/overrides")
        overrides_path.mkdir(parents=True, exist_ok=True)
        overrides_file = overrides_path.joinpath("cfn.patches.yml")
        overrides_file.write_text(templates.get_template("svc/overrides/cfn.patches.yml").render())

    def _get_s3_kms_alias_arns(self, application_name, config):
        application = load_application(application_name, self.session)
        arns = {}

        for environment_name in application.environments:
            kms_provider = self.kms_provider(
                application.environments[environment_name].session.client("kms")
            )

            if environment_name not in config:
                continue

            if "bucket_name" not in config[environment_name]:
                continue

            bucket_name = config[environment_name]["bucket_name"]
            alias_name = f"alias/{application_name}-{environment_name}-{bucket_name}-key"

            try:
                response = kms_provider.describe_key(alias_name)

            # Boto3 classifies all AWS service errors and exceptions as ClientError exceptions
            except botocore.exceptions.ClientError as error:
                if error.response["Error"]["Code"] == "NotFoundException":
                    pass
            else:
                arns[environment_name] = response["KeyMetadata"]["Arn"]

        return arns

    def _get_extensions(self):
        config = self._validate_and_normalise_extensions_config(
            self.PACKAGE_DIR / "default-extensions.yml"
        )
        project_config = self._validate_and_normalise_extensions_config(
            PLATFORM_CONFIG_FILE, "extensions"
        )
        config.update(project_config)
        return config

    def _generate_override_files(self, base_path, file_path, output_dir):
        def generate_files_for_dir(pattern):
            for file in file_path.glob(pattern):
                if file.is_file():
                    contents = file.read_text()
                    file_name = str(file).removeprefix(f"{file_path}/")
                    self.io.info(
                        self.file_provider.mkfile(
                            base_path,
                            output_dir / file_name,
                            contents,
                            overwrite=True,
                        )
                    )

        generate_files_for_dir("*")
        generate_files_for_dir("bin/*")

    def _generate_env_overrides(self, output_dir):
        path = "templates/env/terraform-overrides"
        self.io.info("\n>>> Generating Environment overrides\n")
        overrides_path = output_dir.joinpath(f"copilot/environments/overrides")
        overrides_path.mkdir(parents=True, exist_ok=True)
        template_overrides_path = Path(__file__).parent.parent.joinpath(path)
        self._generate_override_files(Path("."), template_overrides_path, overrides_path)

    def _generate_service_addons(
        self,
        addon_config,
        addon_name,
        addon_type,
        output_dir,
        service_addon_config,
        templates,
        log_destination_arns,
    ):
        # generate svc addons
        for addon_template in ADDON_TEMPLATE_MAP.get(addon_type, []):
            template = templates.get_template(addon_template)

            for svc in addon_config.get("services", []):
                service_path = Path(f"copilot/{svc}/addons/")

                contents = template.render(
                    {
                        "addon_config": service_addon_config,
                        "log_destination": log_destination_arns,
                    }
                )

                (output_dir / service_path).mkdir(parents=True, exist_ok=True)
                self.io.info(
                    self.file_provider.mkfile(
                        output_dir, service_path / f"{addon_name}.yml", contents, overwrite=True
                    )
                )

    def _cleanup_old_files(self, config, output_dir, env_addons_path, env_overrides_path):
        def _rmdir(path):
            if not path.exists():
                return
            for f in path.iterdir():
                if f.is_file():
                    f.unlink()
                if f.is_dir():
                    _rmdir(f)
                    f.rmdir()

        _rmdir(output_dir / env_addons_path)
        _rmdir(output_dir / env_overrides_path)

        all_services = set()
        for services in [v["services"] for v in config.values() if "services" in v]:
            all_services.update(services)

        for service in all_services:
            svc_addons_dir = Path(output_dir, "copilot", service, "addons")
            if not svc_addons_dir.exists():
                continue
            for f in svc_addons_dir.iterdir():
                if f.is_file():
                    f.unlink()



#!/usr/bin/env python
import click

from dbt_platform_helper.domain.pipelines import Pipelines
from dbt_platform_helper.domain.versioning import PlatformHelperVersioning
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.config_validator import ConfigValidator
from dbt_platform_helper.providers.ecr import ECRProvider
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.terraform_manifest import TerraformManifestProvider
from dbt_platform_helper.utils.aws import get_codestar_connection_arn
from dbt_platform_helper.utils.click import ClickDocOptGroup
from dbt_platform_helper.utils.git import git_remote


@click.group(chain=True, cls=ClickDocOptGroup)
def pipeline():
    """Pipeline commands."""
    PlatformHelperVersioning().check_if_needs_update()


@pipeline.command()
@click.option(
    "--deploy-branch",
    help="""Specify the branch of <application>-deploy used to configure the source stage in the environment-pipeline resource. 
    This is generated from the terraform/environments-pipeline/<aws_account>/main.tf file. 
    (Default <application>-deploy branch is specified in 
    <application>-deploy/platform-config.yml/environment_pipelines/<environment-pipeline>/branch).""",
    default=None,
)
def generate(deploy_branch: str):
    """
    Given a platform-config.yml file, generate environment and service
    deployment pipelines.

    This command does the following in relation to the environment pipelines:
    - Reads contents of `platform-config.yml/environment_pipelines` configuration.
      The `terraform/environment-pipelines/<aws_account>/main.tf` file is generated using this configuration.
      The `main.tf` file is then used to generate Terraform for creating an environment pipeline resource.

    This command does the following in relation to the codebase pipelines:
    - Reads contents of `platform-config.yml/codebase_pipelines` configuration.
      The `terraform/codebase-pipelines/main.tf.json` file is generated using this configuration.
      The `main.tf.json` file is then used to generate Terraform for creating a codebase pipeline resource.
    """
    io = ClickIOProvider()
    try:
        pipelines = Pipelines(
            ConfigProvider(ConfigValidator()),
            TerraformManifestProvider(),
            ECRProvider(),
            git_remote,
            get_codestar_connection_arn,
            io,
        )
        pipelines.generate(deploy_branch)
    except Exception as exc:
        io.abort_with_error(str(exc))



import json
import os
import re
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Dict

import boto3

from dbt_platform_helper.constants import PLATFORM_CONFIG_FILE
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.utils.aws import get_aws_session_or_abort
from dbt_platform_helper.utils.aws import get_profile_name_from_account_id
from dbt_platform_helper.utils.aws import get_ssm_secrets
from dbt_platform_helper.utils.messages import abort_with_error


@dataclass
class Environment:
    name: str
    account_id: str
    sessions: Dict[str, boto3.Session]

    @property
    def session(self):
        if self.account_id not in self.sessions:
            self.sessions[self.account_id] = get_aws_session_or_abort(
                get_profile_name_from_account_id(self.account_id),
            )

        return self.sessions[self.account_id]


@dataclass
class Service:
    name: str
    kind: str


@dataclass
class Application:
    name: str
    environments: Dict[str, Environment] = field(default_factory=dict)
    services: Dict[str, Service] = field(default_factory=dict)

    def __str__(self):
        output = f"Application {self.name} with"

        environments = [f"{env.name}:{env.account_id}" for env in self.environments.values()]

        if environments:
            return f"{output} environments {', '.join(environments)}"

        return f"{output} no environments"

    def __eq__(self, other):
        return str(self) == str(other)


def load_application(app=None, default_session=None) -> Application:
    application = Application(app if app else get_application_name())
    current_session = default_session if default_session else get_aws_session_or_abort()

    ssm_client = current_session.client("ssm")

    try:
        ssm_client.get_parameter(
            Name=f"/copilot/applications/{application.name}",
            WithDecryption=False,
        )
    except ssm_client.exceptions.ParameterNotFound:
        raise ApplicationNotFoundException(application.name)

    path = f"/copilot/applications/{application.name}/environments"
    secrets = get_ssm_secrets(app, None, current_session, path)

    sts_client = current_session.client("sts")
    account_id = sts_client.get_caller_identity()["Account"]
    sessions = {account_id: current_session}

    def is_environment_key(name):
        """
        Match only parameter names that are an environment path with no further
        nesting.

        e.g.
         - /copilot/applications/test/environments/my_env will match.
         - /copilot/applications/test/environments/my_env/addons will not match.
        """
        environment_key_regex = r"^/copilot/applications/{}/environments/[^/]*$".format(
            application.name
        )
        return bool(re.match(environment_key_regex, name))

    environments = {
        env["name"]: Environment(env["name"], env["accountID"], sessions)
        for env in [json.loads(s[1]) for s in secrets if is_environment_key(s[0])]
    }
    application.environments = environments

    response = ssm_client.get_parameters_by_path(
        Path=f"/copilot/applications/{application.name}/components",
        Recursive=False,
        WithDecryption=False,
    )
    results = response["Parameters"]
    while "NextToken" in response:
        response = ssm_client.get_parameters_by_path(
            Path=f"/copilot/applications/{application.name}/components",
            Recursive=False,
            WithDecryption=False,
            NextToken=response["NextToken"],
        )
        results.extend(response["Parameters"])

    application.services = {
        svc["name"]: Service(svc["name"], svc["type"])
        for svc in [json.loads(parameter["Value"]) for parameter in results]
    }

    return application


def get_application_name(abort=abort_with_error):
    if Path(PLATFORM_CONFIG_FILE).exists():
        config = ConfigProvider(installed_version_provider="N/A")
        try:
            app_config = config.load_unvalidated_config_file()
            return app_config["application"]
        except KeyError:
            abort(
                f"Cannot get application name. No 'application' key can be found in {PLATFORM_CONFIG_FILE}"
            )
    else:
        abort(f"Cannot get application name. {PLATFORM_CONFIG_FILE} is missing.")


class ApplicationException(PlatformException):
    pass


class ApplicationNotFoundException(ApplicationException):
    def __init__(self, application_name: str):
        super().__init__(
            f"""The account "{os.environ.get("AWS_PROFILE")}" does not contain the application "{application_name}"; ensure you have set the environment variable "AWS_PROFILE" correctly."""
        )


class ApplicationServiceNotFoundException(ApplicationException):
    def __init__(self, application_name: str, svc_name: str):
        super().__init__(
            f"""The service {svc_name} was not found in the application {application_name}. It either does not exist, or has not been deployed."""
        )


class ApplicationEnvironmentNotFoundException(ApplicationException):
    def __init__(self, application_name: str, environment: str):
        super().__init__(
            f"""The environment "{environment}" either does not exist or has not been deployed for the application {application_name}."""
        )



import botocore
from boto3 import Session

from dbt_platform_helper.providers.aws.exceptions import CreateAccessTokenException
from dbt_platform_helper.providers.aws.exceptions import UnableToRetrieveSSOAccountList
from dbt_platform_helper.utils.aws import get_aws_session_or_abort


class SSOAuthProvider:
    def __init__(self, session: Session = None):
        self.session = session
        self.sso_oidc = self._get_client("sso-oidc")
        self.sso = self._get_client("sso")

    def register(self, client_name, client_type):
        client = self.sso_oidc.register_client(clientName=client_name, clientType=client_type)
        client_id = client.get("clientId")
        client_secret = client.get("clientSecret")

        return client_id, client_secret

    def start_device_authorization(self, client_id, client_secret, start_url):
        authz = self.sso_oidc.start_device_authorization(
            clientId=client_id,
            clientSecret=client_secret,
            startUrl=start_url,
        )
        url = authz.get("verificationUriComplete")
        deviceCode = authz.get("deviceCode")

        return url, deviceCode

    def create_access_token(self, client_id, client_secret, device_code):
        try:
            response = self.sso_oidc.create_token(
                clientId=client_id,
                clientSecret=client_secret,
                grantType="urn:ietf:params:oauth:grant-type:device_code",
                deviceCode=device_code,
            )

            return response.get("accessToken")

        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] != "AuthorizationPendingException":
                raise CreateAccessTokenException(client_id)

    def list_accounts(self, access_token, max_results=100):
        aws_accounts_response = self.sso.list_accounts(
            accessToken=access_token,
            maxResults=max_results,
        )

        if len(aws_accounts_response.get("accountList", [])) == 0:
            raise UnableToRetrieveSSOAccountList()
        return aws_accounts_response.get("accountList")

    def _get_client(self, client: str):
        if not self.session:
            self.session = get_aws_session_or_abort()
        return self.session.client(client)



import click

from dbt_platform_helper.domain.notify import Notify
from dbt_platform_helper.domain.versioning import PlatformHelperVersioning
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.slack_channel_notifier import SlackChannelNotifier
from dbt_platform_helper.utils.click import ClickDocOptGroup


@click.group(cls=ClickDocOptGroup, help="Send Slack notifications")
def notify():
    PlatformHelperVersioning().check_if_needs_update()


@notify.command(
    help="Send environment progress notifications. This creates (or updates if --slack-ref is provided) the top level message to the channel."
)
@click.argument("slack-channel-id")
@click.argument("slack-token")
@click.argument("message")
@click.option("--build-arn")
@click.option("--repository")
@click.option("--commit-sha")
@click.option("--slack-ref", help="Slack message reference of the message to update")
def environment_progress(
    slack_channel_id: str,
    slack_token: str,
    message: str,
    build_arn: str,
    repository: str,
    commit_sha: str,
    slack_ref: str,
):
    try:
        io = ClickIOProvider()
        slack_notifier = SlackChannelNotifier(slack_token, slack_channel_id)
        result = Notify(slack_notifier).environment_progress(
            original_message_ref=slack_ref,
            message=message,
            build_arn=build_arn,
            repository=repository,
            commit_sha=commit_sha,
        )

        io.info(result)
    except PlatformException as err:
        io.abort_with_error(str(err))


@notify.command(help="Add a comment to an existing Slack message")
@click.argument("slack-channel-id")
@click.argument("slack-token")
@click.argument("slack-ref")
@click.argument("message")
@click.option("--title", default=None, help="Message title")
@click.option("--send-to-main-channel", default=False, help="Send to main channel")
def add_comment(
    slack_channel_id: str,
    slack_token: str,
    slack_ref: str,
    message: str,
    title: str,
    send_to_main_channel: bool,
):
    try:
        slack_notifier = SlackChannelNotifier(slack_token, slack_channel_id)
        Notify(slack_notifier).add_comment(
            message=message,
            title=title,
            reply_broadcast=send_to_main_channel,
            original_message_ref=slack_ref,
        )
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))
